@@ -18,20 +18,18 @@
  *
  */
 
-#include "xmpp_tasks.h"
-
-//#include <QtCrypto>
-//#include "sha1.h"
-#include "xmpp_xmlcommon.h"
-//#include "xmpp_stream.h"
-//#include "xmpp_types.h"
-#include "xmpp_vcard.h"
-#include "xmpp_bitsofbinary.h"
-
-#include <qregexp.h>
+#include <QRegExp>
 #include <QList>
 #include <QTimer>
 
+#include "xmpp_tasks.h"
+#include "xmpp_xmlcommon.h"
+#include "xmpp_vcard.h"
+#include "xmpp_bitsofbinary.h"
+#include "xmpp_captcha.h"
+#include "xmpp/base/timezone.h"
+#include "xmpp_caps.h"
+
 using namespace XMPP;
 
 
@@ -587,16 +585,11 @@ void JT_Presence::pres(const Status &s)
 			tag.appendChild(x);
 		}
 
-		if(!s.capsNode().isEmpty() && !s.capsVersion().isEmpty()) {
-			QDomElement c = doc()->createElement("c");
-			c.setAttribute("xmlns","http://jabber.org/protocol/caps");
-			if (!s.capsHashAlgorithm().isEmpty())
-				c.setAttribute("hash",s.capsHashAlgorithm());
-			c.setAttribute("node",s.capsNode());
-			c.setAttribute("ver",s.capsVersion());
-			if (!s.capsExt().isEmpty())
-				c.setAttribute("ext",s.capsExt());
-			tag.appendChild(c);
+		if (client()->capsManager()->isEnabled()) {
+			CapsSpec cs = client()->caps();
+			if (cs.isValid()) {
+				tag.appendChild(cs.toXml(doc()));
+			}
 		}
 
 		if(s.isMUC()) {
@@ -761,10 +754,11 @@ bool JT_PushPresence::take(const QDomElement &e)
 		else if(i.tagName() == "x" && i.attribute("xmlns") == "http://jabber.org/protocol/e2e") {
 			p.setKeyID(tagContent(i));
 		}
- 		else if(i.tagName() == "c" && i.attribute("xmlns") == "http://jabber.org/protocol/caps") {
- 			p.setCapsNode(i.attribute("node"));
- 			p.setCapsVersion(i.attribute("ver"));
- 			p.setCapsExt(i.attribute("ext"));
+ 		else if(i.tagName() == "c" && i.attribute("xmlns") == NS_CAPS) {
+			p.setCaps(CapsSpec::fromXml(i));
+			if(!e.hasAttribute("type") && p.caps().isValid()) {
+				client()->capsManager()->updateCaps(j, p.caps());
+			}
   		}
 		else if(i.tagName() == "x" && i.attribute("xmlns") == "vcard-temp:x:update") {
 			QDomElement t;
@@ -852,13 +846,12 @@ static QDomElement oldStyleNS(const QDomElement &e)
 	return i;
 }
 
-JT_Message::JT_Message(Task *parent, const Message &msg, bool want_notify)
+JT_Message::JT_Message(Task *parent, const Message &msg)
 :Task(parent)
 {
 	m = msg;
 	if (m.id().isEmpty())
 		m.setId(id());
-	w_notify = want_notify;
 }
 
 JT_Message::~JT_Message()
@@ -869,7 +862,7 @@ void JT_Message::onGo()
 {
 	Stanza s = m.toStanza(&(client()->stream()));
 	QDomElement e = oldStyleNS(s.element());
-	send(e, w_notify);
+	send(e);
 	setSuccess();
 }
 
@@ -1048,12 +1041,13 @@ void JT_VCard::set(const VCard &card)
 	d->iq.appendChild(card.toXml(doc()) );
 }
 
-void JT_VCard::set(const Jid &j, const VCard &card)
+// isTarget is when we setting target's vcard. for example in case of muc own vcard
+void JT_VCard::set(const Jid &j, const VCard &card, bool isTarget)
 {
 	type = 1;
 	d->vcard = card;
 	d->jid = j;
-	d->iq = createIQ(doc(), "set", "", id());
+	d->iq = createIQ(doc(), "set", isTarget? j.full() : "", id());
 	d->iq.appendChild(card.toXml(doc()) );
 }
 
@@ -1078,7 +1072,8 @@ bool JT_VCard::take(const QDomElement &x)
 					continue;
 
 				if(q.tagName().toUpper() == "VCARD") {
-					if(d->vcard.fromXml(q)) {
+					d->vcard = VCard::fromXml(q);
+					if(d->vcard) {
 						setSuccess();
 						return true;
 					}
@@ -1350,52 +1345,81 @@ const QString & JT_ClientVersion::os() const
 
 
 //----------------------------------------------------------------------------
-// JT_ClientTime
+// JT_EntityTime
 //----------------------------------------------------------------------------
-/*JT_ClientTime::JT_ClientTime(Task *parent, const Jid &_j)
-:Task(parent)
+JT_EntityTime::JT_EntityTime(Task* parent) : Task(parent)
 {
-	j = _j;
-	iq = createIQ("get", j.full(), id());
-	QDomElement query = doc()->createElement("query");
-	query.setAttribute("xmlns", "jabber:iq:time");
-	iq.appendChild(query);
 }
 
-void JT_ClientTime::go()
+/**
+ * \brief Queried entity's JID.
+ */
+const Jid & JT_EntityTime::jid() const
+{
+	return j;
+}
+
+/**
+ * \brief Prepares the task to get information from JID.
+ */
+void JT_EntityTime::get(const Jid &jid)
+{
+	j = jid;
+	iq = createIQ(doc(), "get", jid.full(), id());
+	QDomElement time = doc()->createElement("time");
+	time.setAttribute("xmlns", "urn:xmpp:time");
+	iq.appendChild(time);
+}
+
+void JT_EntityTime::onGo()
 {
 	send(iq);
 }
 
-bool JT_ClientTime::take(const QDomElement &x)
+bool JT_EntityTime::take(const QDomElement &x)
 {
-	if(x.attribute("id") != id())
-		return FALSE;
+	if (!iqVerify(x, j, id()))
+		return false;
 
-	if(x.attribute("type") == "result") {
-		bool found;
-		QDomElement q = queryTag(x);
+	if (x.attribute("type") == "result") {
+		QDomElement q = x.firstChildElement("time");
 		QDomElement tag;
-		tag = findSubTag(q, "utc", &found);
-		if(found)
-			stamp2TS(tagContent(tag), &utc);
-		tag = findSubTag(q, "tz", &found);
-		if(found)
-			timezone = tagContent(tag);
-		tag = findSubTag(q, "display", &found);
-		if(found)
-			display = tagContent(tag);
-
-		setSuccess(TRUE);
+		tag = q.firstChildElement("utc");
+		do {
+			if (tag.isNull()) {
+				break;
+			}
+			utc = QDateTime::fromString(tagContent(tag), Qt::ISODate);
+			tag = q.firstChildElement("tzo");
+			if (!utc.isValid() || tag.isNull()) {
+				break;
+			}
+			tzo = TimeZone::tzdToInt(tagContent(tag));
+			if (tzo == -1) {
+				break;
+			}
+			setSuccess();
+			return true;
+		}
+		while (false);
+		setError(406);
 	}
 	else {
-		setError(getErrorString(x));
-		setSuccess(FALSE);
+		setError(x);
 	}
 
-	return TRUE;
+	return true;
+}
+
+const QDateTime & JT_EntityTime::dateTime() const
+{
+	return utc;
+}
+
+int JT_EntityTime::timezoneOffset() const
+{
+	return tzo;
 }
-*/
 
 
 //----------------------------------------------------------------------------
@@ -1423,129 +1447,22 @@ bool JT_ServInfo::take(const QDomElement &e)
 		iq.appendChild(query);
 		query.appendChild(textTag(doc(), "name", client()->clientName()));
 		query.appendChild(textTag(doc(), "version", client()->clientVersion()));
-		query.appendChild(textTag(doc(), "os", client()->OSName()));
+		query.appendChild(textTag(doc(), "os", client()->OSName() + ' ' + client()->OSVersion()));
 		send(iq);
 		return true;
 	}
-	//else if(ns == "jabber:iq:time") {
-	//	QDomElement iq = createIQ("result", e.attribute("from"), e.attribute("id"));
-	//	QDomElement query = doc()->createElement("query");
-	//	query.setAttribute("xmlns", "jabber:iq:time");
-	//	iq.appendChild(query);
-	//	QDateTime local = QDateTime::currentDateTime();
-	//	QDateTime utc = local.addSecs(-getTZOffset() * 3600);
-	//	QString str = getTZString();
-	//	query.appendChild(textTag("utc", TS2stamp(utc)));
-	//	query.appendChild(textTag("tz", str));
-	//	query.appendChild(textTag("display", QString("%1 %2").arg(local.toString()).arg(str)));
-	//	send(iq);
-	//	return TRUE;
-	//}
 	else if(ns == "http://jabber.org/protocol/disco#info") {
 		// Find out the node
-		bool invalid_node = false;
 		QString node;
 		QDomElement q = e.firstChildElement("query");
 		if(!q.isNull()) // NOTE: Should always be true, since a NS was found above
 			node = q.attribute("node");
 
-		QDomElement iq = createIQ(doc(), "result", e.attribute("from"), e.attribute("id"));
-		QDomElement query = doc()->createElement("query");
-		query.setAttribute("xmlns", "http://jabber.org/protocol/disco#info");
-		if (!node.isEmpty())
-			query.setAttribute("node", node);
-		iq.appendChild(query);
+		if (node.isEmpty() || node == client()->caps().flatten()) {
 
-		// Identity
-		DiscoItem::Identity identity = client()->identity();
-		QDomElement id = doc()->createElement("identity");
-		if (!identity.category.isEmpty() && !identity.type.isEmpty()) {
-			id.setAttribute("category",identity.category);
-			id.setAttribute("type",identity.type);
-			if (!identity.name.isEmpty()) {
-				id.setAttribute("name",identity.name);
-			}
-		}
-		else {
-			// Default values
-			id.setAttribute("category","client");
-			id.setAttribute("type","pc");
-		}
-		query.appendChild(id);
-
-		QDomElement feature;
-		if (node.isEmpty() || node == client()->capsNode() + "#" + client()->capsVersion()) {
-			if (client()->fileTransferManager()) {
-				// Standard features
-				feature = doc()->createElement("feature");
-				feature.setAttribute("var", "http://jabber.org/protocol/bytestreams");
-				query.appendChild(feature);
-
-				feature = doc()->createElement("feature");
-				feature.setAttribute("var", "http://jabber.org/protocol/ibb");
-				query.appendChild(feature);
-
-				feature = doc()->createElement("feature");
-				feature.setAttribute("var", "http://jabber.org/protocol/si");
-				query.appendChild(feature);
-
-				feature = doc()->createElement("feature");
-				feature.setAttribute("var", "http://jabber.org/protocol/si/profile/file-transfer");
-				query.appendChild(feature);
-			}
-
-			feature = doc()->createElement("feature");
-			feature.setAttribute("var", "http://jabber.org/protocol/disco#info");
-			query.appendChild(feature);
-
-			feature = doc()->createElement("feature");
-			feature.setAttribute("var", "urn:xmpp:bob");
-			query.appendChild(feature);
-
-			feature = doc()->createElement("feature");
-			feature.setAttribute("var", "urn:xmpp:ping");
-			query.appendChild(feature);
-
-			// Client-specific features
-			QStringList clientFeatures = client()->features().list();
-			foreach (const QString & i, clientFeatures) {
-				feature = doc()->createElement("feature");
-				feature.setAttribute("var", i);
-				query.appendChild(feature);
-			}
-
-			if (node.isEmpty()) {
-				// Extended features
-				const QStringList exts = client()->extensions();
-				for (QStringList::ConstIterator i = exts.begin(); i != exts.end(); ++i) {
-					const QStringList& l = client()->extension(*i).list();
-					for ( QStringList::ConstIterator j = l.begin(); j != l.end(); ++j ) {
-						feature = doc()->createElement("feature");
-						feature.setAttribute("var", *j);
-						query.appendChild(feature);
-					}
-				}
-			}
-		}
-		else if (node.startsWith(client()->capsNode() + "#")) {
-			QString ext = node.right(node.length()-client()->capsNode().length()-1);
-			if (client()->extensions().contains(ext)) {
-				const QStringList& l = client()->extension(ext).list();
-				for ( QStringList::ConstIterator it = l.begin(); it != l.end(); ++it ) {
-					feature = doc()->createElement("feature");
-					feature.setAttribute("var", *it);
-					query.appendChild(feature);
-				}
-			}
-			else {
-				invalid_node = true;
-			}
-		}
-		else {
-			invalid_node = true;
-		}
-
-		if (!invalid_node) {
+			QDomElement iq = createIQ(doc(), "result", e.attribute("from"), e.attribute("id"));
+			DiscoItem item = client()->makeDiscoResult(node);
+			iq.appendChild(item.toDiscoInfoResult(doc()));
 			send(iq);
 		}
 		else {
@@ -1568,6 +1485,31 @@ bool JT_ServInfo::take(const QDomElement &e)
 		}
 		return true;
 	}
+	if (!ns.isEmpty()) {
+		return false;
+	}
+
+	ns = e.firstChildElement("time").attribute("xmlns");
+	if (ns == "urn:xmpp:time") {
+		QDomElement iq = createIQ(doc(), "result", e.attribute("from"), e.attribute("id"));
+		QDomElement time = doc()->createElement("time");
+		time.setAttribute("xmlns", ns);
+		iq.appendChild(time);
+
+		QDateTime local = QDateTime::currentDateTime();
+
+		int off = TimeZone::offsetFromUtc();
+		QTime t = QTime(0, 0).addSecs(qAbs(off)*60);
+		QString tzo = (off < 0 ? "-" : "+") + t.toString("HH:mm");
+		time.appendChild(textTag(doc(), "tzo", tzo));
+		QString localTimeStr = local.toUTC().toString(Qt::ISODate);
+		if (!localTimeStr.endsWith("Z"))
+			localTimeStr.append("Z");
+		time.appendChild(textTag(doc(), "utc", localTimeStr));
+
+		send(iq);
+		return true;
+	}
 
 	return false;
 }
@@ -2067,7 +2009,7 @@ bool JT_BitsOfBinary::take(const QDomElement &x)
 	return true;
 }
 
-BoBData JT_BitsOfBinary::data()
+BoBData &JT_BitsOfBinary::data()
 {
 	return d->data;
 }
@@ -2099,3 +2041,142 @@ bool JT_PongServer::take(const QDomElement &e)
 	}
 	return false;
 }
+
+//---------------------------------------------------------------------------
+// JT_CaptchaChallenger
+//---------------------------------------------------------------------------
+class JT_CaptchaChallenger::Private
+{
+public:
+	Jid j;
+	CaptchaChallenge challenge;
+};
+
+JT_CaptchaChallenger::JT_CaptchaChallenger(Task *parent) :
+    Task(parent),
+    d(new Private)
+{
+}
+
+JT_CaptchaChallenger::~JT_CaptchaChallenger()
+{
+	delete d;
+}
+
+void JT_CaptchaChallenger::set(const Jid &j, const CaptchaChallenge &c)
+{
+	d->j = j;
+	d->challenge = c;
+}
+
+void JT_CaptchaChallenger::onGo()
+{
+	setTimeout(CaptchaValidTimeout);
+
+	Message m;
+	m.setId(id());
+	m.setBody(d->challenge.explanation());
+	m.setUrlList(d->challenge.urls());
+
+	XData form = d->challenge.form();
+	XData::FieldList fl = form.fields();
+	XData::FieldList::Iterator it;
+	for (it = fl.begin(); it < fl.end(); ++it) {
+		if (it->var() == "challenge" && it->type() == XData::Field::Field_Hidden) {
+			it->setValue(QStringList() << id());
+		}
+	}
+	if (it == fl.end()) {
+		XData::Field f;
+		f.setType(XData::Field::Field_Hidden);
+		f.setVar("challenge");
+		f.setValue(QStringList() << id());
+		fl.append(f);
+	}
+	form.setFields(fl);
+
+	m.setForm(form);
+	m.setTo(d->j);
+	client()->sendMessage(m);
+}
+
+bool JT_CaptchaChallenger::take(const QDomElement &x)
+{
+	if(x.tagName() == "message" && x.attribute("id") == id() &&
+	        Jid(x.attribute("from")) == d->j && !x.firstChildElement("error").isNull())
+	{
+		setError(x);
+		return true;
+	}
+
+	XDomNodeList nl;
+	XData xd;
+	QString rid = x.attribute("id");
+	if (rid.isEmpty() || x.tagName() != "iq" ||
+	        Jid(x.attribute("from")) != d->j || x.attribute("type") != "set" ||
+	        (nl = childElementsByTagNameNS(x, "urn:xmpp:captcha", "captcha")).isEmpty() ||
+	        (nl = childElementsByTagNameNS(nl.item(0).toElement(), "jabber:x:data", "x")).isEmpty() ||
+	        (xd.fromXml(nl.item(0).toElement()), xd.getField("challenge").value().value(0) != id()))
+	{
+		return false;
+	}
+
+	CaptchaChallenge::Result r = d->challenge.validateResponse(xd);
+	QDomElement iq;
+	if (r == CaptchaChallenge::Passed) {
+		iq = createIQ(doc(), "result", d->j.full(), rid);
+	} else {
+		Stanza::Error::ErrorCond ec;
+		if (r == CaptchaChallenge::Unavailable) {
+			ec = Stanza::Error::ServiceUnavailable;
+		} else {
+			ec = Stanza::Error::NotAcceptable;
+		}
+		iq = createIQ(doc(), "error", d->j.full(), rid);
+		Stanza::Error error(Stanza::Error::Cancel, ec);
+		iq.appendChild(error.toXml(*doc(), client()->stream().baseNS()));
+	}
+	send(iq);
+
+	setSuccess();
+
+	return true;
+}
+
+
+//---------------------------------------------------------------------------
+// JT_CaptchaSender
+//---------------------------------------------------------------------------
+JT_CaptchaSender::JT_CaptchaSender(Task *parent) :
+    Task(parent)
+{}
+
+void JT_CaptchaSender::set(const Jid &j, const XData &xd)
+{
+	to = j;
+
+	iq = createIQ(doc(), "set", to.full(), id());
+	iq.appendChild(doc()->createElementNS("urn:xmpp:captcha", "captcha"))
+	        .appendChild(xd.toXml(doc(), true));
+}
+
+void JT_CaptchaSender::onGo()
+{
+	send(iq);
+}
+
+bool JT_CaptchaSender::take(const QDomElement &x)
+{
+	if (!iqVerify(x, to, id())) {
+		return false;
+	}
+
+	if (x.attribute("type") == "result") {
+		setSuccess();
+	}
+	else {
+		setError(x);
+	}
+
+	return true;
+}
